PUB bootFile(p) | c                   ' Start Spin program from file
'  p     = Address of file control block as described earlier
'  returns false if the file is not found.  Does not return if successful.
'          This routine stops all cogs except the one executing this call,
'          then starts a loader cog and stops itself leaving only the loader.
'          This loader reads in the Spin/Assembly program stored in the file
'          using the header stored in the 1st 16 bytes of the program to
'          provide the size of the program to be loaded, the starting address
'          of the program's stack area, and the size and location of the
'          program's variable area.  These are used to initialize the stack
'          and the variable area.  The clock is initialized as necessary and
'          a new Spin interpreter is started in the cog used by the loader.
   bootDelay := clkfreq / 100         ' 10ms delay in clock ticks
   long[p][2] &= $00FFFFFF            ' Set current segment to $00
   if findFile(p)                     ' Specified file must be present
      repeat c from 0 to 7
         if c <> cogid                ' Stop all cogs other than the one
            cogstop(c)                '  running this bootFile execution
      if cognew(@entryPoint,long[p][3]>>10)
         cogstop(cogid)               ' Stop current cog while boot proceeds

DAT
'' Boot loader for Winbond flash memory.  This is passed the starting block
'' number of the file to be loaded (in 4K blocks).  A short delay allows the
'' caller to stop itself.  This reads each block in the file into memory up
'' through the end of the program as determined by pbase * 4.  Once the program
'' is loaded, the stack is initialized, variable space is zeroed, the required
'' system clock is initialized, and a Spin interpreter is started in the same
'' cog as the loader.  If an error occurs, the Propeller will likely hang.

                        org     0
entryPoint              mov     flashTemp,bootDelay     ' Delay for 10ms to
                        add     flashTemp,CNT           '  allow caller to
                        waitcnt flashTemp,#0            '   stop itself
                        mov     flashNext,PAR           ' Get starting block #
                        shl     flashNext,#10
                        mov     loadAddr,#0             ' Start loading at 0
readFlash               andn    outa,diraMask           ' All output pins low
                        or      dira,diraMask           ' DOpin is only input
                        movd    :headerAcc,#header      ' Get sector header
                        mov     flashCnt,#4             '  into 4 long area
:bitLoop6               mov     flashMask,#$80
:bitLoop7               mov     bitCount,#8             ' Get a byte at a time
:bitLoop8               test    DOMask,ina           wz
                        or      outa,ClkMask
                        andn    outa,ClkMask
:headerAcc              muxnz   0-0,flashMask
                        ror     flashMask,#1
                        djnz    bitCount,#:bitLoop8     ' Go back for next bit
                        rol     flashMask,#8         wc ' Reposition bit mask
                if_nc   jmp     #:bitLoop7              ' Done when shifts out
                        add     :headerAcc,incrDest     ' On to next long
                        djnz    flashCnt,#:bitLoop6
                        mov     flashCnt,dataSize       ' Size of data area
:bitLoop4               mov     flashMask,#$80          ' Switch to reading
:bitLoop5               test    DOMask,ina           wz '  data area to hub
                        or      outa,ClkMask
                        andn    outa,ClkMask
                        muxnz   flashTemp,flashMask
                        shr     flashMask,#1         wz ' Reposition bit mask
                if_nc   jmp     #:bitLoop5              ' Done when shifts out
                        wrbyte  flashTemp,loadAddr      ' Write the byte to hub
                        add     loadAddr,#1
                        djnz    flashCnt,#:bitLoop4
                        mov     flashNext,header+3      ' Get link to next blk
                        and     flashNext,wordMask
                        cmp     flashNext,wordMask   wz ' End of file?
                if_nz   shl     flashNext,#12           ' Go on to next block
                if_nz   or      outa,CSMask             ' End of transaction
                if_nz   andn    dira,diraMask
                if_nz   jmp     #readFlash

startUpSpin             rdword  i2cTemp,#$000A          ' Get dbase value
                        sub     i2cTemp,#4
                        wrlong  StackMark,i2cTemp       ' Stack marker @ dbase
                        sub     i2cTemp,#4
                        wrlong  StackMark,i2cTemp
                        rdword  i2cOther,#$0008         ' Get vbase value
                        sub     i2cTemp,i2cOther
                        shr     i2cTemp,#2           wz ' # of longs between
:zeroIt        if_nz    wrlong  i2cZero,i2cOther        '  vbase and below
               if_nz    add     i2cOther,#4             '   stack marker
               if_nz    djnz    i2cTemp,#:zeroIt        ' Zero that space
                        rdlong  i2cTemp,#$0000
                        cmp     i2cTemp,SaveClkFreq  wz ' Clock freqs same?
                        rdbyte  i2cTemp,#$0004          ' Clock modes same?
               if_ne    jmp     #:changeClock
                        cmp     i2cTemp,SaveClkMode  wz ' Same?  Go start COG
                if_e    jmp     #:justStartUp
:changeClock            and     i2cTemp,#$F8            ' Use RCFAST clock while
                        clkset  i2cTemp                 '  clock starts up
                        mov     i2cTemp,time_xtal
:startupDelay           djnz    i2cTemp,#:startupDelay  ' Allow 20ms @20MHz for
                        rdbyte  i2cTemp,#$0004          '  xtal/pll to settle
                        clkset  i2cTemp                 ' Switch to selected clk
:justStartUp            cogid   i2cOther                ' Use this cog for SPIN
                        or      i2cOther,interpreter
                        coginit i2cOther

readFlash               andn    outa,diraMask           ' All output pins low
                        or      dira,diraMask           ' DOpin is only input
                        mov     flashMask,msbMask
:shiftMask              test    flashTemp,flashMask  wz ' Find first nonzero
                if_z    shr     flashMask,#8            '  byte from MSB
                if_z    tjnz    flashMask,#:shiftMask
                        and     flashMask,msbitMask     ' Set up bitmask
:bitLoop1               test    flashTemp,flashMask  wz
                        muxnz   outa,DIOMask            ' Output from MSBit to
                        or      outa,ClkMask            '  LSBit then clock
                        andn    outa,ClkMask
                        shr     flashMask,#1            ' Shift bit mask
                        tjnz    flashMask,#:bitLoop1
                        test    flashCnt,wordSign    wz ' Is count < 0
                if_z    jmp     #:skipOutput
                        xor     flashCnt,wordMask       ' Abs value - 1
                        add     flashCnt,#1
:bitLoop2               mov     flashMask,#$80
                        rdbyte  flashTemp,flashAddr
:bitLoop3               test    flashTemp,flashMask  wz
                        muxnz   outa,DIOMask
                        or      outa,ClkMask
                        andn    outa,ClkMask
                        shr     flashMask,#1
                        tjnz    flashMask,#:bitLoop3
                        add     flashAddr,#1
                        djnz    flashCnt,#:bitLoop2
:skipOutput             tjz     flashCnt,#:skipInput
:bitLoop4               mov     flashMask,#$80
:bitLoop5               test    DOMask,ina           wz
                        or      outa,ClkMask
                        andn    outa,ClkMask
                        muxnz   flashTemp,flashMask
                        shr     flashMask,#1
                        tjnz    flashMask,#:bitLoop5
                        wrbyte  flashTemp,flashAddr
                        add     flashAddr,#1
                        djnz    flashCnt,#:bitLoop4
:skipInput              or      outa,CSMask
                        andn    dira,diraMask
                        jmp     #newOpFetch

bootDelay               long    0                       ' 10ms delay time
time_xtal               long    20 * 20000 / 4 / 1      ' 20ms (@20MHz, DJNZ)
interpreter             long    ($0004 << 16) | ($F004 << 2) | %0000
StackMark               long    $FFF9FFFF               ' Mark stack base x 2
noLinkPtr               long    $FFFF << 12             ' Null link
endOfMemory             long    $7FF0                   ' Last 16 RAM bytes
linkWord                long    $7FFC                   ' Link to next block
loadSize                long    dataSize                ' Number of data bytes
incrDest                long    $200
wordSign                long    $0000_8000
msbitMask               long    $80_80_80_80
msbMask                 long    $FF000000
addrMask                long    $00FFFFFF
wordMask                long    $0000FFFF
diraMask                long    |<CSpin + |<DIOpin + |<Clkpin
DIOMask                 long    |<DIOpin
ClkMask                 long    |<Clkpin
CSMask                  long    |<CSpin

loadAddr                res     1                       ' Hub address for load
flashAddr               res     1                       ' Flash memory address
hubAddr                 res     1                       ' Hub address for read
flashCnt                res     1                       ' # of bytes to read
flashTemp               res     1
flashLoc                res     1
flashMask               res     1
flashChar               res     1

header                  res     4                       ' 4K block header